import numpy as np
from photutils import centroid_com
import ipywidgets as ipw

from astropy.stats import sigma_clipped_stats
from astropy.table import Table
from astropy.nddata import Cutout2D

import matplotlib.pyplot as plt
from stellarphot.visualization import seeing_plot

__all__ = ['set_keybindings', 'box', 'make_show_event']

out = ipw.Output()
out2 = ipw.Output()
out3 = ipw.Output()


def set_keybindings(image_widget, scroll_zoom=False):
    """
    Set image widget keyboard bindings. The bindings are:

    + Pan by click-and-drag or with arrow keys.
    + Zoom by scrolling or using the ``+``/``-`` keys.
    + Adjust contrast by Ctrl-left click and drag; reset with
      shift-right-click.

    Any existing key bindings are removed.

    Parameters
    ----------

    image_widget : astrowidgets.ImageWidget
        Image widget on which to set the key bindings.
    """
    bind_map = image_widget._viewer.get_bindmap()
    # Displays the event map...
    # bind_map.eventmap
    bind_map.clear_event_map()
    bind_map.map_event(None, (), 'ms_left', 'pan')
    if scroll_zoom:
        bind_map.map_event(None, (), 'pa_pan', 'zoom')

    # bind_map.map_event(None, (), 'ms_left', 'cursor')
    # contrast with right mouse
    bind_map.map_event(None, (), 'ms_right', 'contrast')

    # shift-right mouse to reset contrast
    bind_map.map_event(None, ('shift',), 'ms_right', 'contrast_restore')
    bind_map.map_event(None, ('ctrl',), 'ms_left', 'cursor')

    # Bind +/- to zoom in/out
    bind_map.map_event(None, (), 'kp_+', 'zoom_in')
    bind_map.map_event(None, (), 'kp_=', 'zoom_in')
    bind_map.map_event(None, (), 'kp_-', 'zoom_out')
    bind_map.map_event(None, (), 'kp__', 'zoom_out')

    # Bind arrow keys to panning
    # There is NOT a typo below. I want the keys to move the image in the
    # direction of the arrow
    bind_map.map_event(None, (), 'kp_left', 'pan_right')
    bind_map.map_event(None, (), 'kp_right', 'pan_left')
    bind_map.map_event(None, (), 'kp_up', 'pan_down')
    bind_map.map_event(None, (), 'kp_down', 'pan_up')


def find_center(image, center_guess, cutout_size=30, max_iters=10):
    """
    Find the centroid of a star from an initial guess of its position. Originally
    written to find star from a mouse click.

    Parameters
    ----------

    image : numpy array or CCDData
        Image containing the star.

    center_guess : array or tuple
        The position, in pixels, of the initial guess for the position of
        the star. The coordinates should be horizontal first, then vertical,
        i.e. opposite the usual Python convention for a numpy array.

    cutout_size : int, optional
        The default width of the cutout to use for finding the star.

    max_iters : int, optional
        Maximum number of iterations to go through in finding the center.
    """
    pad = cutout_size // 2
    x, y = center_guess

    # Keep track of iterations
    cnt = 0

    # Grab the cutout...
    sub_data = image[y - pad:y + pad, x - pad:x + pad]  # - med

    # ...do stats on it...
    _, sub_med, _ = sigma_clipped_stats(sub_data)
    # sub_med = 0

    # ...and centroid.
    x_cm, y_cm = centroid_com(sub_data - sub_med)

    # Translate centroid back to original image (maybe use Cutout2D instead)
    cen = np.array([x_cm + x - pad, y_cm + y - pad])

    # ceno is the "original" center guess, set it to something nonsensical here
    ceno = np.array([-100, -100])

    while (cnt <= max_iters and
           (np.abs(np.array([x_cm, y_cm]) - pad).max() > 3
            or np.abs(cen - ceno).max() > 0.1)):

        # Update x, y positions for subsetting
        x = int(np.floor(x_cm)) + x - pad
        y = int(np.floor(y_cm)) + y - pad
        sub_data = image[y - pad:y + pad, x - pad:x + pad]  # - med
        _, sub_med, _ = sigma_clipped_stats(sub_data)
        # sub_med = 0
        mask = (sub_data - sub_med) < 0
        x_cm, y_cm = centroid_com(sub_data - sub_med, mask=mask)
        ceno = cen
        cen = np.array([x_cm + x - pad, y_cm + y - pad])
        if not np.all(~np.isnan(cen)):
            raise RuntimeError('Centroid finding failed, '
                               'previous was {}, current is {}'.format(ceno, cen))
        cnt += 1

    return cen


def radial_profile(data, center, size=30, return_scaled=True):
    """
    Construct a radial profile of a chunk of width ``size`` centered
    at ``center`` from image ``data`.

    Parameters
    ----------

    data : numpy array or CCDData
        Image data
    center : list-like
        x, y position of the center in pixel coordinates, i.e. horizontal
        coordinate then vertical.
    size : int, optional
        Width of the rectangular cutout to use in constructing the profile.
    return_scaled : bool, optional
        If ``True``, return an average radius and profile, otherwise
        it is cumulative. Not at all clear what a "cumulative" radius
        means, tbh.

    Returns
    -------

    r_exact : numpy array
        Exact radius of center of each pixels from profile center.
    ravg : numpy array
        Average radius of pixels in each bin.
    radialprofile : numpy array
        Radial profile.
    """
    yd, xd = np.indices((size, size))

    sub_image = Cutout2D(data, center, size, mode='strict')
    sub_center = sub_image.center_cutout

    r = np.sqrt((xd - sub_center[0])**2 + (yd - sub_center[1])**2)
    r_exact = r.copy()
    r = r.astype(np.int)

    sub_data = sub_image.data

    tbin = np.bincount(r.ravel(), sub_data.ravel())
    rbin = np.bincount(r.ravel(), r_exact.ravel())
    nr = np.bincount(r.ravel())
    if return_scaled:
        radialprofile = tbin / nr
        ravg = rbin / nr
    else:
        radialprofile = tbin
        ravg = rbin

    return r_exact, ravg, radialprofile


def find_hwhm(r, intensity):
    """
    Estimate HWHM from normalized, angle-averaged intensity profile.
    """
    # Make the bold assumption that intensity decreases monotonically
    # so that we just need to find the first place where intensity is
    # less than 0.5 to estimate the HWHM.
    less_than_half = intensity < 0.5
    half_index = np.arange(len(less_than_half))[less_than_half][0]
    before_half = half_index - 1

    # Do linear interpolation to find the radius at which the intensity
    # is 0.5.
    r_more = r[before_half]
    r_less = r[half_index]
    I_more = intensity[before_half]
    I_less = intensity[half_index]

    I_half = 0.5

    r_half = r_less - (I_less - I_half) / (I_less - I_more) * (r_less - r_more)

    return r_half


def make_show_event(iw):
    def show_event(viewer, event, datax, datay):

        i = iw._viewer.get_image()
        data = i.get_data()

        # Rough location of click in original image
        x = int(np.floor(event.data_x))
        y = int(np.floor(event.data_y))

        cen = find_center(data, (x, y), cutout_size=30)

        # ADD MARKER WHERE CLICKED
        iw.add_markers(Table(data=[[cen[0]], [cen[1]]], names=['x', 'y']))
        # print(cen[0], cen[1])

        # ----> MOVE PROFILE CONSTRUCTION INTO FUNCTION <----

        # CONSTRUCT RADIAL PROFILE OF PATCH AROUND STAR
        profile_size = 40
        r_exact, ravg, radialprofile = radial_profile(data, cen, size=profile_size)
        sub_data = Cutout2D(data, cen, size=profile_size).data
        sub_med = np.median(sub_data)
        adjust_max = radialprofile.max() - sub_med
        scaled_profile = (radialprofile - sub_med) / adjust_max
        scaled_exact_counts = (sub_data - sub_med) / adjust_max

        # DISPLAY THE SCALED PROFILE
        out.clear_output(wait=True)
        with out:
            # print(dir(event))
            # print(event.data_x, event.data_y)
            plt.clf()
            # sub_med += med
            HWHM = find_hwhm(ravg, scaled_profile)
            seeing_plot(r_exact, scaled_exact_counts, ravg, scaled_profile, HWHM,
                        'Some Image Name', file_name='some_name', gap=6, annulus_width=13)
            plt.show()

        # CALCULATE AND DISPLAY NET COUNTS INSIDE RADIUS
        out2.clear_output(wait=True)
        with out2:
            r_exact, ravg, tbin2 = radial_profile(data - sub_med, cen,
                                                  size=profile_size,
                                                  return_scaled=False)
            #tbin2 = np.bincount(r.ravel(), (sub_data - sub_med).ravel())
            counts = np.cumsum(tbin2)

            plt.plot(range(len(radialprofile)), counts)
            plt.xlim(0, 20)
            plt.grid()

            sub_blot = sub_data.copy()
            sub_blot[10:30, 10:30] = np.nan
            sub_std = np.nanstd(sub_blot)
            plt.title('Net counts in aperture std {:.2f} med {:.2f}'.format(
                sub_std, sub_med))
            e_sky = np.sqrt(sub_med)
            plt.xlabel('Aperture radius')
            plt.show()

        # CALCULATE And DISPLAY SNR AS A FUNCTION OF RADIUS
        out3.clear_output(wait=True)
        with out3:
            rn = 10
            poisson = np.sqrt(np.cumsum(tbin2))
            nr = tbin2 / radialprofile
            error = np.sqrt(poisson ** 2 + np.cumsum(nr)
                            * (e_sky ** 2 + rn ** 2))
            snr = np.cumsum(tbin2) / error
            # snr_max = snr[:20].max()
            plt.plot(range(len(radialprofile)), snr)
            plt.title('Signal to noise ratio {}'.format(snr.max()))
            plt.xlim(0, 20)
            # plt.ylim(0, 2)
            plt.xlabel('Aperture radius')
            plt.grid()
            plt.show()
    return show_event


def box(imagewidget):
    big_box = ipw.HBox()
    big_box = ipw.GridspecLayout(1, 2)
    layout = ipw.Layout(width='20ch')
    hb = ipw.HBox()
    ap_t = ipw.IntText(description='Aperture', value=5, layout=layout)
    in_t = ipw.IntText(description='Inner annulus', value=10, layout=layout)
    out_t = ipw.IntText(description='Outer annulus', value=20, layout=layout)
    hb.children = [ap_t, in_t, out_t]

    lil_box = ipw.VBox()
    lil_box.children = [out, out2, out3]
    # big_box.children = [imagewidget, lil_box]
    big_box[0, 0] = imagewidget
    big_box[0, 1] = lil_box
    big_box.layout.width = '100%'
    # Line below puts space between the image and the plots so the plots
    # don't jump around as the image value changes.
    big_box.layout.justify_content = 'space-between'

    return big_box



from pathlib import Path
import re

import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table, hstack

import numpy as np

__all__ = ['parse_aij_table', 'ApertureFileAIJ']


class ApertureAIJ:
    """
    Represent the aperture information AstroImageJ saves.
    """
    def __init__(self):
        # Outer annulus radius
        self.rback2 = 41.0

        # Inner annulus radius
        self.rback1 = 27.0

        # Aperture radius
        self.radius = 15.0

        # Defaults for this match the defaults for stellarphot.
        self.removebackstars = True

        # Not sure what this does but stellarphot doesn't use it.
        self.backplane = False


class MultiApertureAIJ:
    """
    Represent the multi-aperture information that AstroImageJ saves
    """
    def __init__(self):
        # Default values for these chosen to match AIJ defaults
        # They are not used by stellarphot
        self.naperturesmax = 1000
        self.apfwhmfactor = 1.4
        self.usevarsizeap = False

        # Each attribute below should be list-like, all of the same length
        self.isrefstar = []
        self.centroidstar = []
        self.isalignstar = []

        self.xapertures = []
        self.yapertures = []

        self.absmagapertures = []

        self.raapertures = []
        self.decapertures = []


class ApertureFileAIJ:
    """
    Represent AstroImageJ aperture file.
    """
    def __init__(self):
        self.aperture = ApertureAIJ()
        self.multiaperture = MultiApertureAIJ()

    def __str__(self):
        lines = []

        top_level_attrib = vars(self)
        for name, attrib in top_level_attrib.items():
            base_attrib = vars(attrib)
            for bname, battrib in base_attrib.items():
                try:
                    value = ','.join([str(v).lower() for v in battrib])
                except TypeError:
                    value = battrib

                if value is True:
                    value = 'true'
                elif value is False:
                    value = 'false'

                lines.append(f'.{name}.{bname}={value}')

        # Add a trailing blank line
        return '\n'.join(lines) + '\n'

    def write(self, file):
        p = Path(file)
        p.write_text(str(self))

    @classmethod
    def from_table(cls, aperture_table,
                   aperture_rad=None, inner_annulus=None, outer_annulus=None,
                   default_absmag=99.999, default_isalign=True,
                   default_centroidstar=True):
        """
        Create an `stellarphot.io.ApertureFileAIJ` from a stellarphot aperture
        table and info about the aperture sizes.

        Parameters
        ----------

        aperture_table : `astropy.table.Table`
            Table of aperture information.

        aperture_rad : number
            Radius of aperture.

        inner_annulus : number
            Inner radius of annulus.

        outer_annulus : number
            Outer radius of annulus.
        """
        # Create the instance
        apAIJ = cls()

        # Populate aperture properties
        apAIJ.aperture.rback2 = outer_annulus
        apAIJ.aperture.rback1 = inner_annulus
        apAIJ.aperture.radius = aperture_rad

        # Remaining properties of apAIJ.aperture default to the
        # correct values for stellarphot

        # Populate multiaperture properties
        n_apertures = len(aperture_table)
        columns = aperture_table.colnames

        # A boolean column for this would be better, but this will do
        # for now.
        apAIJ.multiaperture.isrefstar = [('comparison' in name.lower())
                           for name in aperture_table['marker name']]

        # These are not currently in the table but that could change...
        if 'centroidstar' not in columns:
            apAIJ.multiaperture.centroidstar = [default_centroidstar] * n_apertures
        else:
            apAIJ.multiaperture.centroidstar = aperture_table['centroidstar']

        if 'isalign' not in columns:
            apAIJ.multiaperture.isalignstar = [default_isalign] * n_apertures
        else:
            apAIJ.multiaperture.isalignstar = aperture_table['isalign']

        apAIJ.multiaperture.xapertures = aperture_table['x']
        apAIJ.multiaperture.yapertures = aperture_table['y']

        if 'absmag' not in columns:
            apAIJ.multiaperture.absmagapertures = [default_absmag] * n_apertures
        else:
            apAIJ.multiaperture.absmagapertures = aperture_table['absmag']

        apAIJ.multiaperture.raapertures = aperture_table['coord'].ra.degree
        apAIJ.multiaperture.decapertures = aperture_table['coord'].dec.degree

        return apAIJ


def _is_comp(star_coord, comp_table):
    idx, d2d, _ = star_coord.match_to_catalog_sky(comp_table['coord'])
    return 'comparison' in comp_table['marker name'][idx]


def generate_aij_table(table_name, comparison_table):
    info_columns = {
        'date-obs': 'DATE_OBS',
        'airmass': 'AIRMASS',
        'BJD': 'BJD_MOBS',
        'exposure': 'EXPOSURE',
        'filter': 'FILTER',
        'aperture': 'Source_Radius',
        'annulus_inner': 'Sky_Rad(min)',
        'annulus_outer': 'Sky_Rad(max)'
    }
    by_source_columns = {
        'xcenter': 'X(IJ)',
        'ycenter': 'Y(IJ)',
        'aperture_net_flux': 'Source-Sky',
        'aperture_area': 'N_Src_Pixels',
        'noise-aij': 'Source_Error',
        'snr': 'Source_SNR',
        'sky_per_pix_avg': 'Sky/Pixel',
        'annulus_area': 'N_Sky_Pixels',
        'fwhm_x': 'X-Width',
        'fwhm_y': 'Y-Width',
        'width': 'Width',
        'relative_flux': 'rel_flux',
        'relative_flux_error': 'rel_flux_err',
        'relative_flux_snr': 'rel_flux_SNR',
        'comparison counts': 'tot_C_cnts',
        'comparison error': 'tot_C_err'
    }
    by_star = table_name.group_by('id')
    base_table = by_star.groups[0][list(info_columns.keys())]
    for star_id, sub_table in zip(by_star.groups.keys, by_star.groups):
        star_co = SkyCoord(ra=sub_table['RA'][0], dec=sub_table['Dec'][0],
                           unit='degree')

        if _is_comp(star_co, comparison_table):
            char = 'C'
        else:
            char = 'T'

        new_table = sub_table[list(by_source_columns.keys())] #  + ['BJD']

        for old_col, new_col in by_source_columns.items():
            new_column_name = new_col + f'_{char}{star_id[0]}'
            new_table.rename_column(old_col, new_column_name)
        base_table = hstack([base_table, new_table])
    for old_col, new_col in info_columns.items():
        base_table.rename_column(old_col, new_col)

    return base_table

def parse_aij_table(table_name):
    """
    Return a list of objects, one for each source, from an AstroImageJ
    photometry table.

    Parameters
    ----------

    table_name : str
        Name of the table.
    """

    # Read in the raw table.
    if table_name.endswith('.csv'):
        # The table may have been edited and changed to csv.
        raw = Table.read(table_name)
    else:
        # The default, though, is tab-separated text with a file extension xls.
        raw = Table.read(table_name, format='ascii.tab')

    # Extract the names of all columns which are not specific to a source.
    # Source columns end with _TX or _CX where X is one or more digits.
    source_column = r'.*_[CT]\d+'
    common_columns = []
    for name in raw.colnames:
        if not re.search(source_column, name):
            if name.strip():
                common_columns.append(name)

    # Get all of the source designations from the names of the net counts
    # columns.
    flux_columns = [name for name in raw.colnames if
                    name.startswith('Source-Sky')]
    source_column_ids = [c.split('_')[1] for c in flux_columns]

    # For the first source, grab all of the column names that are specific to
    # a source. These will be the same for all sources.
    first_source_names = [name for name in raw.colnames if
                          name.endswith(source_column_ids[0])]
    generic_source_columns = [name.rsplit('_', 1)[0] for name in
                              first_source_names]

    # Make a list of star objects. Not sure this is actually better than
    # a list of tables or something simple like that.
    stars = []
    for idx, source in enumerate(source_column_ids):
        specifc_column_names = ['_'.join([g, source]) for g in
                                generic_source_columns]
        all_names = common_columns + specifc_column_names
        my_table = raw[all_names]
        for spec, gen in zip(specifc_column_names, generic_source_columns):
            my_table.rename_column(spec, gen)
        stars.append(Star(my_table, idx + 1))

    return stars


class Star(object):
    def __init__(self, table, id_num):
        self._table = table
        self._table['DEC'].unit = u.degree
        self.id = id_num

    @property
    def airmass(self):
        return self._table['AIRMASS']

    @property
    def counts(self):
        return self._table['Source-Sky']

    @property
    def ra(self):
        return self._table['RA'] / 24 * 360 * u.degree

    @property
    def dec(self):
        return self._table['DEC']

    @property
    def error(self):
        return self._table['Source_Error']

    @property
    def sky_per_pixel(self):
        return self._table['Sky/Pixel']

    @property
    def peak(self):
        return self._table['Peak']

    @property
    def jd_utc_start(self):
        return self._table['JD_UTC']

    @property
    def mjd_start(self):
        return self._table['J.D.-2400000'] - 0.5

    @property
    def exposure(self):
        try:
            return self._table['EXPOSURE']
        except KeyError:
            return self._table['EXPTIME']

    @property
    def magnitude(self):
        return -2.5 * np.log10(self.counts) + 2.5 * np.log10(self.exposure)

    @property
    def snr(self):
        return self.counts / self.error

    @property
    def magnitude_error(self):
        return 1.0857 / self.snr

    @property
    def bjd_tdb(self):
        return self._table['BJD_TDB']



from pathlib import Path

import pandas

import ipywidgets as ipw

import numpy as np

from astropy.table import Table
from astropy.coordinates import SkyCoord
from astropy import units as u
from astropy.nddata import CCDData
from astrowidgets import ImageWidget

from stellarphot.differential_photometry import *
from stellarphot.photometry import *
from stellarphot.visualization.seeing_profile_functions import set_keybindings


__all__ = ['read_file', 'set_up', 'match', 'mag_scale',
           'in_field', 'make_markers', 'wrap']


def read_file(radec_file):
    """
    Read an AIJ radec file with target and/or comp positions

    Parameters
    ----------

    radec_file : str
        Name of the file

    Returns
    -------

    `astropy.table.Table`
        Table with target information, including a
        `astropy.coordinates.SkyCoord` column.
    """
    df = pandas.read_csv(radec_file, names=['RA', 'Dec', 'a', 'b', 'Mag'])
    target_table = Table.from_pandas(df)
    ra = target_table['RA']
    dec = target_table['Dec']
    target_table['coords'] = SkyCoord(ra=ra, dec=dec, unit=(u.hour, u.degree))
    return target_table


def set_up(sample_image_for_finding_stars,
           directory_with_images='.'):
    """
    Find known variable in the field of view and read sample image.

    Parameters
    ----------

    sample_image_for_finding_stars : str
        Name or URL of a fits image of the field of view.

    directory_with_images : str, optional
        Folder in which the image is located. Ignored if the sample image
        is a URL.
    """
    if sample_image_for_finding_stars.startswith('http'):
        path = sample_image_for_finding_stars
    else:
        path = Path(directory_with_images) / sample_image_for_finding_stars

    ccd = CCDData.read(path)
    vsx = find_known_variables(ccd)
    ra = vsx['RAJ2000']
    dec = vsx['DEJ2000']
    vsx['coords'] = SkyCoord(ra=ra, dec=dec, unit=(u.hour, u.degree))

    return ccd, vsx


def match(CCD, RD, vsx):
    """
    Find APASS stars in FOV and matches APASS stars to VSX and
    APASS to input targets.
    """
    apass, apass_in_bright = find_apass_stars(
        CCD)
    ra = apass['RAJ2000']
    dec = apass['DEJ2000']
    apass['coords'] = SkyCoord(ra=ra, dec=dec, unit=(u.hour, u.degree))
    apass_coord = apass['coords']

    if vsx:
        v_index, v_angle, v_dist = \
            apass_coord.match_to_catalog_sky(vsx['coords'])
    else:
        v_angle = []
    if RD:
        RD_index, RD_angle, RD_dist = \
            apass_coord.match_to_catalog_sky(RD['coords'])
    else:
        RD_angle = []
    return apass, v_angle, RD_angle


def mag_scale(cmag, apass, v_angle, RD_angle,
              brighter_dmag=0.44, dimmer_dmag=0.75):
    """
    Select comparison stars that are 1) not close the VSX stars or to other
    target stars and 2) fall within a particular magnitude range.


    """
    high_mag = apass['r_mag'] < cmag + dimmer_dmag
    low_mag = apass['r_mag'] > cmag - brighter_dmag
    if v_angle:
        good_v_angle = v_angle > 1.0 * u.arcsec
    else:
        good_v_angle = True

    if RD_angle:
        good_RD_angle = RD_angle > 1.0 * u.arcsec
    else:
        good_RD_angle = True

    good_stars = high_mag & low_mag & good_RD_angle & good_v_angle
    good_apass = apass[good_stars]
    apass_good_coord = good_apass['coords']
    return apass_good_coord, good_stars


def in_field(apass_good_coord, ccd, apass, good_stars):
    """
    Return apass stars in the field of view
    """
    apassx, apassy = ccd.wcs.all_world2pix(
        apass_good_coord.ra, apass_good_coord.dec, 0)
    ccdx, ccdy = ccd.shape

    xin = (apassx < ccdx) & (0 < apassx)
    yin = (apassy < ccdy) & (0 < apassy)
    xy_in = xin & yin
    apass_good_coord[xy_in]
    nt = apass[good_stars]
    ent = nt[xy_in]
    return ent


def make_markers(iw, ccd, RD, vsx, ent,
                 name_or_coord=None):
    """
    Add markers for APASS, TESS targets, VSX.
    Also center on object/coordinate.
    """
    iw.load_nddata(ccd)
    iw.zoom_level = 'fit'
    iw.reset_markers()

    if RD:
        iw.marker = {'type': 'circle', 'color': 'green', 'radius': 10}
        iw.add_markers(RD, skycoord_colname='coords',
                       use_skycoord=True, marker_name='TESS Targets')

    if name_or_coord is not None:
        if isinstance(name_or_coord, str):
            iw.center_on(SkyCoord.from_name(name_or_coord))
        else:
            iw.center_on(name_or_coord)

    iw.marker = {'type': 'circle', 'color': 'blue', 'radius': 10}
    iw.add_markers(vsx, skycoord_colname='coords',
                   use_skycoord=True, marker_name='VSX')

    iw.marker = {'type': 'circle', 'color': 'red', 'radius': 10}
    iw.add_markers(ent, skycoord_colname='coords',
                   use_skycoord=True, marker_name='APASS comparison')
    iw.marker = {'type': 'cross', 'color': 'red', 'radius': 6}


def wrap(imagewidget, outputwidget):
    """
    Make the bits that let you click to select/deselect comparisons
    """
    def cb(viewer, event, data_x, data_y):
        i = imagewidget._viewer.get_image()

        try:
            imagewidget.next_elim += 1
        except AttributeError:
            imagewidget.next_elim = 1
        pad = 15
        x = int(np.floor(event.data_x))
        y = int(np.floor(event.data_y))
        ra, dec = i.wcs.wcs.all_pix2world(event.data_x, event.data_y, 0)
        out_skycoord = SkyCoord(ra=ra, dec=dec, unit=(u.degree, u.degree))
        all_table = imagewidget.get_markers(marker_name='all')
        with outputwidget:
            index, d2d, d3d = out_skycoord.match_to_catalog_sky(
                all_table['coord'])
            if d2d < 10 * u.arcsec:
                mouse = all_table['coord'][index].separation(
                    all_table['coord'])
                rat = mouse < 1 * u.arcsec
                elims = [name for name in all_table['marker name']
                         [rat] if name.startswith('elim')]
                if not elims:
                    imagewidget.add_markers(all_table[rat], skycoord_colname='coord', use_skycoord=True,
                                            marker_name=f'elim{imagewidget.next_elim}')
                else:
                    for elim in elims:
                        imagewidget.remove_markers(marker_name=elim)
            else:
                print('sorry try again')
                imagewidget._viewer.onscreen_message('Click closer to a star')
            print(all_table['marker name'][index])
            print(x, y, ra, dec, out_skycoord)
    return cb


def viewer():
    header = ipw.HTML(value="""
    <h2>Click and drag or use arrow keys to pan, use +/- keys to zoom</h2>
    <h3>Shift-left click (or Crtl-left click)to exclude star as target or comp. Click again to include.</h3>
    """)

    legend = ipw.HTML(value="""
    <h3>Green circles -- Gaia stars within 2.5 arcmin of target</h3>
    <h3>Red circles -- APASS stars within 1 mag of target</h3>
    <h3>Blue circles -- VSX variables</h3>
    <h3>Red × -- Exclude as target or comp</h3>
    """)

    iw = ImageWidget()
    out = ipw.Output()
    set_keybindings(iw)
    bind_map = iw._viewer.get_bindmap()
    gvc = iw._viewer.get_canvas()
    bind_map.map_event(None, ('shift',), 'ms_left', 'cursor')
    gvc.add_callback('cursor-down', wrap(iw, out))


    box = ipw.VBox()
    inner_box = ipw.HBox()
    inner_box.children = [iw, legend]
    box.children = [header, inner_box]

    return box, iw
